"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NttExecutorRoute = exports.nttExecutorRoute = void 0;
const sdk_connect_1 = require("@wormhole-foundation/sdk-connect");
require("@wormhole-foundation/sdk-definitions-ntt");
const types_js_1 = require("../types.js");
const utils_js_1 = require("./utils.js");
const sdk_definitions_1 = require("@wormhole-foundation/sdk-definitions");
function nttExecutorRoute(config) {
    class NttExecutorRouteImpl extends NttExecutorRoute {
        static config = config;
    }
    return NttExecutorRouteImpl;
}
exports.nttExecutorRoute = nttExecutorRoute;
class NttExecutorRoute extends sdk_connect_1.routes.AutomaticRoute {
    // executor supports gas drop-off
    static NATIVE_GAS_DROPOFF_SUPPORTED = true;
    // @ts-ignore
    // Since we set the config on the static class, access it with this param
    // the NttExecutorRoute.config will always be empty
    staticConfig = this.constructor.config;
    static config = { ntt: { tokens: {} } };
    static meta = { name: "NttExecutorRoute" };
    static supportedNetworks() {
        return types_js_1.NttRoute.resolveSupportedNetworks(this.config.ntt);
    }
    static supportedChains(network) {
        return types_js_1.NttRoute.resolveSupportedChains(this.config.ntt, network);
    }
    static async supportedSourceTokens(fromChain) {
        return types_js_1.NttRoute.resolveSourceTokens(this.config.ntt, fromChain);
    }
    static async supportedDestinationTokens(sourceToken, fromChain, toChain) {
        return types_js_1.NttRoute.resolveDestinationTokens(this.config.ntt, sourceToken, fromChain, toChain);
    }
    static isProtocolSupported(chain) {
        return chain.supportsProtocol("Ntt");
    }
    getDefaultOptions() {
        return {
            nativeGas: 0,
        };
    }
    async validate(request, params) {
        const options = params.options ?? this.getDefaultOptions();
        if (options.nativeGas !== undefined &&
            (options.nativeGas < 0 || options.nativeGas > 1)) {
            return {
                valid: false,
                error: new Error("Invalid native gas percentage"),
                params,
            };
        }
        const parsedAmount = sdk_connect_1.amount.parse(params.amount, request.source.decimals);
        // IMPORTANT: The EVM NttManager will revert if there is dust. This is not the case for Solana,
        // but we want to be consistent across chains.
        const trimmedAmount = types_js_1.NttRoute.trimAmount(parsedAmount, request.destination.decimals);
        const { srcContracts, dstContracts } = types_js_1.NttRoute.resolveNttContracts(this.staticConfig.ntt, request.source.id, request.destination.id);
        let referrerFeeDbps = 0n;
        if (this.staticConfig.referrerFee) {
            referrerFeeDbps = this.staticConfig.referrerFee.feeDbps;
            if (this.staticConfig.referrerFee.perTokenOverrides) {
                const srcTokenAddress = (0, sdk_connect_1.canonicalAddress)(request.source.id);
                const override = this.staticConfig.referrerFee.perTokenOverrides[request.source.id.chain]?.[srcTokenAddress];
                if (override) {
                    referrerFeeDbps = override.referrerFeeDbps;
                }
            }
        }
        const validatedParams = {
            amount: params.amount,
            normalizedParams: {
                amount: trimmedAmount,
                sourceContracts: srcContracts,
                destinationContracts: dstContracts,
                referrerFeeDbps,
            },
            options,
        };
        return { valid: true, params: validatedParams };
    }
    async quote(request, params) {
        const { fromChain, toChain } = request;
        try {
            const executorQuote = await this.fetchExecutorQuote(request, params);
            const { remainingAmount, estimatedCost, gasDropOff, expires } = executorQuote;
            const receivedAmount = sdk_connect_1.amount.scale(sdk_connect_1.amount.fromBaseUnits(remainingAmount, request.source.decimals), request.destination.decimals);
            const result = {
                success: true,
                params,
                sourceToken: {
                    token: request.source.id,
                    amount: params.normalizedParams.amount,
                },
                destinationToken: {
                    token: request.destination.id,
                    amount: receivedAmount,
                },
                relayFee: {
                    token: (0, sdk_connect_1.nativeTokenId)(fromChain.chain),
                    amount: sdk_connect_1.amount.fromBaseUnits(estimatedCost, fromChain.config.nativeTokenDecimals),
                },
                destinationNativeGas: sdk_connect_1.amount.fromBaseUnits(gasDropOff, toChain.config.nativeTokenDecimals),
                eta: sdk_connect_1.finality.estimateFinalityTime(request.fromChain.chain) +
                    sdk_connect_1.guardians.guardianAttestationEta * 1000,
                expires,
                details: executorQuote,
            };
            const dstNtt = await toChain.getProtocol("Ntt", {
                ntt: params.normalizedParams.destinationContracts,
            });
            const duration = await dstNtt.getRateLimitDuration();
            if (duration > 0n) {
                const capacity = await dstNtt.getCurrentInboundCapacity(fromChain.chain);
                if (types_js_1.NttRoute.isCapacityThresholdExceeded(sdk_connect_1.amount.units(receivedAmount), capacity)) {
                    result.warnings = [
                        {
                            type: "DestinationCapacityWarning",
                            delayDurationSec: Number(duration),
                        },
                    ];
                }
            }
            return result;
        }
        catch (e) {
            return {
                success: false,
                error: e instanceof Error ? e : new Error(String(e)),
            };
        }
    }
    async fetchExecutorQuote(request, params) {
        const { fromChain, toChain } = request;
        let referrer = undefined;
        const referrerFeeConfig = this.staticConfig.referrerFee;
        if (referrerFeeConfig && referrerFeeConfig.feeDbps > 0n) {
            const platform = (0, sdk_connect_1.chainToPlatform)(fromChain.chain);
            const referrerAddress = referrerFeeConfig.referrerAddresses?.[platform] ?? "";
            if (!referrerAddress) {
                throw new Error("No referrer address found");
            }
            referrer = sdk_connect_1.Wormhole.chainAddress(fromChain.chain, referrerAddress);
        }
        const { referrerFee, remainingAmount, referrerFeeDbps } = (0, utils_js_1.calculateReferrerFee)(params.normalizedParams.amount, params.normalizedParams.referrerFeeDbps, request.destination.decimals);
        if (remainingAmount <= 0n) {
            throw new Error("Amount after fee <= 0");
        }
        const capabilities = await (0, utils_js_1.fetchCapabilities)(fromChain.network);
        const srcCapabilities = capabilities[(0, sdk_connect_1.toChainId)(fromChain.chain)];
        if (!srcCapabilities) {
            throw new Error("Unsupported source chain");
        }
        const dstCapabilities = capabilities[(0, sdk_connect_1.toChainId)(toChain.chain)];
        if (!dstCapabilities || !dstCapabilities.requestPrefixes.includes("ERN1")) {
            throw new Error("Unsupported destination chain");
        }
        const { recipient } = request;
        const dstNttWithExec = await toChain.getProtocol("NttWithExecutor", {
            ntt: params.normalizedParams.destinationContracts,
        });
        // Calculate the gas dropOff value
        const gasDropOffLimit = BigInt(dstCapabilities.gasDropOffLimit);
        const dropOff = params.options.nativeGas && gasDropOffLimit > 0n
            ? (BigInt(Math.round(params.options.nativeGas * 100)) *
                gasDropOffLimit) /
                100n
            : 0n;
        const { msgValue, gasLimit } = await dstNttWithExec.estimateMsgValueAndGasLimit(recipient);
        const relayRequests = [];
        // Add the gas instruction
        relayRequests.push({
            request: {
                type: "GasInstruction",
                gasLimit,
                msgValue,
            },
        });
        // Add the gas drop-off instruction if applicable
        if (dropOff > 0n) {
            relayRequests.push({
                request: {
                    type: "GasDropOffInstruction",
                    dropOff,
                    // If the recipient is undefined (e.g. the user hasn’t connected their wallet yet),
                    // we temporarily use a dummy address to fetch a quote.
                    // The recipient address is validated later in the `initiate` method, which will throw if it's still missing.
                    recipient: recipient
                        ? recipient.address.toUniversalAddress()
                        : new sdk_connect_1.UniversalAddress(new Uint8Array(32)),
                },
            });
        }
        const relayInstructions = (0, sdk_connect_1.serializeLayout)(sdk_definitions_1.relayInstructionsLayout, {
            requests: relayRequests,
        });
        const quote = await (0, utils_js_1.fetchSignedQuote)(fromChain.network, fromChain.chain, toChain.chain, sdk_connect_1.encoding.hex.encode(relayInstructions, true));
        if (!quote.estimatedCost) {
            throw new Error("No estimated cost");
        }
        const estimatedCost = BigInt(quote.estimatedCost);
        const signedQuoteBytes = sdk_connect_1.encoding.hex.decode(quote.signedQuote);
        const signedQuote = (0, sdk_connect_1.deserializeLayout)(sdk_definitions_1.signedQuoteLayout, signedQuoteBytes);
        return {
            signedQuote: signedQuoteBytes,
            relayInstructions: relayInstructions,
            estimatedCost,
            payeeAddress: signedQuote.quote.payeeAddress,
            referrer,
            referrerFee,
            remainingAmount,
            referrerFeeDbps,
            expires: signedQuote.quote.expiryTime,
            gasDropOff: dropOff,
        };
    }
    async initiate(request, signer, quote, to) {
        if (!quote.details) {
            throw new Error("Missing quote details");
        }
        const { params, details } = quote;
        const relayInstructions = (0, sdk_connect_1.deserializeLayout)(sdk_definitions_1.relayInstructionsLayout, details.relayInstructions);
        // Make sure that the gas drop-off recipient matches the actual recipient
        relayInstructions.requests.forEach(({ request }) => {
            if (request.type === "GasDropOffInstruction" &&
                !request.recipient.equals(to.address.toUniversalAddress())) {
                throw new Error("Gas drop-off recipient does not match");
            }
        });
        const { fromChain } = request;
        const sender = sdk_connect_1.Wormhole.parseAddress(signer.chain(), signer.address());
        const nttWithExec = await fromChain.getProtocol("NttWithExecutor", {
            ntt: params.normalizedParams.sourceContracts,
        });
        const ntt = await fromChain.getProtocol("Ntt", {
            ntt: params.normalizedParams.sourceContracts,
        });
        const wrapNative = (0, sdk_definitions_1.isNative)(request.source.id.address);
        const initXfer = nttWithExec.transfer(sender, to, sdk_connect_1.amount.units(params.normalizedParams.amount), details, ntt, wrapNative);
        const txids = await (0, sdk_connect_1.signSendWait)(fromChain, initXfer, signer);
        // Status the transfer immediately before returning
        let statusAttempts = 0;
        const statusTransferImmediately = async () => {
            while (statusAttempts < 20) {
                try {
                    const [txStatus] = await (0, utils_js_1.fetchStatus)(fromChain.network, txids.at(-1).txid, fromChain.chain);
                    if (txStatus) {
                        break;
                    }
                }
                catch (_) {
                    // is ok we just try again!
                }
                statusAttempts++;
                await new Promise((resolve) => setTimeout(resolve, 2000));
            }
        };
        // Spawn a loop in the background that will status this transfer until
        // the API gives a successful response. We don't await the result
        // here because we don't need it for the return value.
        statusTransferImmediately();
        return {
            from: fromChain.chain,
            to: to.chain,
            state: sdk_connect_1.TransferState.SourceInitiated,
            originTxs: txids,
            params,
        };
    }
    async complete(signer, receipt) {
        if (!(0, sdk_connect_1.isAttested)(receipt) && !(0, sdk_connect_1.isFailed)(receipt)) {
            if ((0, sdk_connect_1.isRedeemed)(receipt))
                return receipt;
            throw new Error("The source must be finalized in order to complete the transfer");
        }
        if (!receipt.attestation) {
            throw new Error("No attestation found for the transfer");
        }
        const toChain = this.wh.getChain(receipt.to);
        const ntt = await toChain.getProtocol("Ntt", {
            ntt: receipt.params.normalizedParams.destinationContracts,
        });
        const sender = sdk_connect_1.Wormhole.parseAddress(signer.chain(), signer.address());
        const completeXfer = ntt.redeem([receipt.attestation.attestation], sender);
        const txids = await (0, sdk_connect_1.signSendWait)(toChain, completeXfer, signer);
        return {
            ...receipt,
            state: sdk_connect_1.TransferState.DestinationInitiated,
            attestation: receipt.attestation,
            destinationTxs: txids,
        };
    }
    async resume(tx) {
        const vaa = await this.wh.getVaa(tx.txid, "Ntt:WormholeTransfer");
        if (!vaa)
            throw new Error("No VAA found for transaction: " + tx.txid);
        const msgId = {
            chain: vaa.emitterChain,
            emitter: vaa.emitterAddress,
            sequence: vaa.sequence,
        };
        const { recipientChain, trimmedAmount } = vaa.payload["nttManagerPayload"].payload;
        const token = (0, sdk_connect_1.canonicalAddress)({
            chain: vaa.emitterChain,
            address: vaa.payload["nttManagerPayload"].payload.sourceToken,
        });
        const manager = (0, sdk_connect_1.canonicalAddress)({
            chain: vaa.emitterChain,
            address: vaa.payload["sourceNttManager"],
        });
        const whTransceiver = vaa.emitterChain === "Solana"
            ? manager
            : (0, sdk_connect_1.canonicalAddress)({
                chain: vaa.emitterChain,
                address: vaa.emitterAddress,
            });
        const dstInfo = types_js_1.NttRoute.resolveDestinationNttContracts(this.staticConfig.ntt, {
            chain: vaa.emitterChain,
            address: vaa.payload["sourceNttManager"],
        }, recipientChain);
        const amt = sdk_connect_1.amount.fromBaseUnits(trimmedAmount.amount, trimmedAmount.decimals);
        return {
            from: vaa.emitterChain,
            to: recipientChain,
            state: sdk_connect_1.TransferState.Attested,
            originTxs: [tx],
            attestation: {
                id: msgId,
                attestation: vaa,
            },
            params: {
                amount: sdk_connect_1.amount.display(amt),
                options: {},
                normalizedParams: {
                    amount: amt,
                    sourceContracts: {
                        token,
                        manager,
                        transceiver: {
                            wormhole: whTransceiver,
                        },
                    },
                    destinationContracts: {
                        token: dstInfo.token,
                        manager: dstInfo.manager,
                        transceiver: {
                            wormhole: dstInfo.transceiver["wormhole"],
                        },
                    },
                    referrerFeeDbps: 0n,
                },
            },
        };
    }
    // Even though this is an automatic route, the transfer may need to be
    // manually finalized if it was queued
    async finalize(signer, receipt) {
        if (!(0, sdk_connect_1.isDestinationQueued)(receipt)) {
            throw new Error("The transfer must be destination queued in order to finalize");
        }
        const { attestation: { attestation: vaa }, } = receipt;
        const toChain = this.wh.getChain(receipt.to);
        const ntt = await toChain.getProtocol("Ntt", {
            ntt: receipt.params.normalizedParams.destinationContracts,
        });
        const sender = sdk_connect_1.Wormhole.chainAddress(signer.chain(), signer.address());
        const completeTransfer = ntt.completeInboundQueuedTransfer(receipt.from, vaa.payload["nttManagerPayload"], sender.address);
        const finalizeTxids = await (0, sdk_connect_1.signSendWait)(toChain, completeTransfer, signer);
        return {
            ...receipt,
            state: sdk_connect_1.TransferState.DestinationFinalized,
            destinationTxs: [...(receipt.destinationTxs ?? []), ...finalizeTxids],
        };
    }
    async *track(receipt, timeout) {
        // First we fetch the attestation (VAA) for the transfer
        if ((0, sdk_connect_1.isSourceInitiated)(receipt) || (0, sdk_connect_1.isSourceFinalized)(receipt)) {
            const { txid } = receipt.originTxs.at(-1);
            const vaa = await this.wh.getVaa(txid, "Ntt:WormholeTransfer", timeout);
            if (!vaa)
                throw new Error("No VAA found for transaction: " + txid);
            const msgId = {
                chain: vaa.emitterChain,
                emitter: vaa.emitterAddress,
                sequence: vaa.sequence,
            };
            receipt = {
                ...receipt,
                state: sdk_connect_1.TransferState.Attested,
                attestation: {
                    id: msgId,
                    attestation: vaa,
                },
            };
            yield receipt;
        }
        const toChain = this.wh.getChain(receipt.to);
        const ntt = await toChain.getProtocol("Ntt", {
            ntt: receipt.params.normalizedParams.destinationContracts,
        });
        // Check if the relay was successful or failed
        if ((0, sdk_connect_1.isAttested)(receipt) && !(0, sdk_connect_1.isFailed)(receipt)) {
            const [txStatus] = await (0, utils_js_1.fetchStatus)(this.wh.network, receipt.originTxs.at(-1).txid, receipt.from);
            if (!txStatus)
                throw new Error("No transaction status found");
            const relayStatus = txStatus.status;
            if (relayStatus === utils_js_1.RelayStatus.Failed || // this could happen if simulation fails
                relayStatus === utils_js_1.RelayStatus.Underpaid || // only happens if you don't pay at least the costEstimate
                relayStatus === utils_js_1.RelayStatus.Unsupported || // capabilities check didn't pass
                relayStatus === utils_js_1.RelayStatus.Aborted // An unrecoverable error indicating the attempt should stop (bad data, pre-flight checks failed, or chain-specific conditions)
            ) {
                receipt = {
                    ...receipt,
                    state: sdk_connect_1.TransferState.Failed,
                    error: new sdk_connect_1.routes.RelayFailedError(`Relay failed with status: ${relayStatus}`),
                };
                yield receipt;
            }
        }
        // Check if the transfer was redeemed
        if ((0, sdk_connect_1.isAttested)(receipt) || (0, sdk_connect_1.isFailed)(receipt)) {
            if (!receipt.attestation) {
                throw new Error("No attestation found");
            }
            const { attestation: { attestation: vaa }, } = receipt;
            if (await ntt.getIsApproved(vaa)) {
                receipt = {
                    ...receipt,
                    state: sdk_connect_1.TransferState.DestinationInitiated,
                    attestation: receipt.attestation,
                    // TODO: check for destination event transactions to get dest Txids
                };
                yield receipt;
            }
        }
        if ((0, sdk_connect_1.isRedeemed)(receipt) || (0, sdk_connect_1.isDestinationQueued)(receipt)) {
            const { attestation: { attestation: vaa }, } = receipt;
            const queuedTransfer = await ntt.getInboundQueuedTransfer(vaa.emitterChain, vaa.payload["nttManagerPayload"]);
            if (queuedTransfer !== null) {
                receipt = {
                    ...receipt,
                    state: sdk_connect_1.TransferState.DestinationQueued,
                    queueReleaseTime: new Date(queuedTransfer.rateLimitExpiryTimestamp * 1000),
                };
                yield receipt;
            }
            else if (await ntt.getIsExecuted(vaa)) {
                receipt = {
                    ...receipt,
                    state: sdk_connect_1.TransferState.DestinationFinalized,
                };
                yield receipt;
            }
        }
        yield receipt;
    }
}
exports.NttExecutorRoute = NttExecutorRoute;
//# sourceMappingURL=executor.js.map